use reqwest::Client;
use crate::settings::{ApiClientSettings, Settings};
use std::time::Duration;
use crate::models::{ServerInfo, SuccessResponse, Board, ThreadSlice, WarpSlice, Post, NewPost, RequestBody, UpdatedHiddenByOp, ErrorResponse, Auth, AuthInfo, NewUser};
use crate::constants::COMPATIBLE_API_VERSIONS;
use url::Url;
use uuid::Uuid;
use crate::parsers::Markup;
use std::str;
use std::convert::TryFrom;
use serde::de;

pub struct ApiClient {
    pub client: Client,
    pub api_url: Url,
}

type ApiResult<T> = Result<Result<SuccessResponse<T>, ErrorResponse>, Box<dyn std::error::Error>>;

impl ApiClient {
    pub fn new(settings: &Settings) -> Result<Self, Box<dyn std::error::Error>> {
        Ok(ApiClient {
            api_url: Self::get_api_url(&settings.api_server.url)?,
            client: Self::build_client(&settings.api_client)?,
        })
    }

    /// Connect to the server, check if its version is compatible
    pub async fn connect(&self) -> Result<u32, Box<dyn std::error::Error>> {
        let compatible_versions = semver::VersionReq::parse(&COMPATIBLE_API_VERSIONS)?;

        let res = self.client.get(self.api_url.join("serverInfo")?)
            .send().await?;

        let body = res.text().await?;

        let SuccessResponse {
            data: ServerInfo { version, pow_bits },
            pagination: _,
            meta: _,
        } = serde_json::from_str(&body)?;

        let version = semver::Version::parse(&version)?;

        if !compatible_versions.matches(&version) {
            return Err(format!(
                "Server {} has incompatible version: {}. Expected: {}.",
                &self.api_url,
                version,
                compatible_versions,
            ).into());
        }
        Ok(pow_bits)
    }

    pub async fn get_boards(&self, skip: u32, take: u32)
                            -> Result<SuccessResponse<Vec<Board>>, Box<dyn std::error::Error>>
    {
        let res = self.client.get(self.api_url.join("boards")?)
            .query(&[
                ("skip", skip.to_string()),
                ("take", take.to_string()),
                ("includeTotal", "true".to_string()),
            ])
            .send().await?;

        let body = res.text().await?;

        let res: SuccessResponse<Vec<Board>> = serde_json::from_str(&body)?;
        Ok(res)
    }

    pub async fn get_thread_slice(&self, thread_id: &Uuid, skip: u32, take: u32,
                                  last_slice: bool, db: &sled::Db)
                                  -> Result<SuccessResponse<ThreadSlice>, Box<dyn std::error::Error>>
    {
        let skip = skip + 1; // to skip OP

        let res = self.client.get(
            self.api_url.join("threadSlices/")?.join(&thread_id.to_string())?
        ).query(&[
            ("skip", skip.to_string()),
            ("take", take.to_string()),
            ("lastSlice", last_slice.to_string()),
            ("includeBoard", "true".to_string()),
        ]).send().await?;

        let body = res.text().await?;

        let mut res: SuccessResponse<ThreadSlice> = serde_json::from_str(&body)?;

        res.data.op.text = Self::get_post_text_html(&res.data.op, db)?;

        for post in res.data.posts.iter_mut() {
            post.text = Self::get_post_text_html(post, db)?;
        }

        Ok(res)
    }

    pub async fn get_warp_slice(&self, board_name: Option<&String>, skip: u32, take: u32, db: &sled::Db)
                                -> Result<SuccessResponse<WarpSlice>, Box<dyn std::error::Error>>
    {
        let include_board = board_name.is_none();

        let mut query_params: Vec<(&str, String)> = vec![
            ("skip", skip.to_string()),
            ("take", take.to_string()),
            ("includeBoard", include_board.to_string()),
        ];

        if let Some(board_name) = board_name {
            query_params.push(("board", board_name.to_owned()));
        }

        let res = self.client.get(self.api_url.join("warpSlices/")?)
            .query(&query_params)
            .send().await?;

        let body = res.text().await?;

        let mut res: SuccessResponse<WarpSlice> = serde_json::from_str(&body)?;

        for st in res.data.shell_threads.iter_mut() {
            let omitted_count = st.thread.post_count
                .saturating_sub(u32::try_from(st.posts.len())?)
                .saturating_sub(1); // OP

            st.omitted_count = if omitted_count == 0 { None } else { Some(omitted_count) };

            st.op.text = Self::get_post_text_html(&st.op, db)?;

            for post in st.posts.iter_mut() {
                post.text = Self::get_post_text_html(post, db)?;
            };
        }

        Ok(res)
    }

    pub async fn get_post(&self, post_id: &Uuid)
                          -> Result<SuccessResponse<Post>, Box<dyn std::error::Error>>
    {
        let res = self.client.get(
            self.api_url.join("posts/")?.join(&post_id.to_string())?
        ).query(&[
            ("includeIndex", "true"),
            ("includeThread", "true"),
            ("includeBoard", "true"),
        ]).send().await?;

        let body = res.text().await?;

        let res: SuccessResponse<Post> = serde_json::from_str(&body)?;
        Ok(res)
    }

    pub async fn get_recent_posts(&self, board: Option<&str>, db: &sled::Db)
                                  -> Result<SuccessResponse<Vec<Post>>, Box<dyn std::error::Error>>
    {
        let res = self.client.get(self.api_url.join("posts/")?).query(&[
            ("sortByDate", Some("desc")),
            ("includeThread", Some("true")),
            ("includeBoard", Some("true")),
            ("board", board),
        ]).send().await?;

        let body = res.text().await?;

        let mut res: SuccessResponse<Vec<Post>> = serde_json::from_str(&body)?;

        for post in res.data.iter_mut() {
            post.text = Self::get_post_text_html(post, db)?;
        };

        Ok(res)
    }

    pub async fn create_post(&self, new_post: NewPost, access_token: Option<&String>)
                             -> ApiResult<Post>
    {
        let request_body = RequestBody { data: new_post };

        let res = Self::add_auth_header(
            self.client.post(self.api_url.join("posts")?),
            access_token,
        ).json(&request_body).send().await?;

        let body = res.text().await?;

        Self::get_api_result(&body)
    }

    pub async fn update_board(&self, board: Board, access_token: Option<&String>) -> ApiResult<String>
    {
        let request_body = RequestBody { data: board };

        let res = Self::add_auth_header(
            self.client.patch(self.api_url.join("boards")?),
            access_token,
        ).json(&request_body).send().await?;

        let body = res.text().await?;

        Self::get_api_result(&body)
    }

    pub async fn authorize(&self, auth: &Auth) -> ApiResult<AuthInfo>
    {
        let mut auth = auth.clone();
        auth.back_url = None;

        let request_body = RequestBody { data: auth };

        let res = self.client.post(self.api_url.join("authorize")?)
            .json(&request_body)
            .send().await?;

        let body = res.text().await?;

        Self::get_api_result(&body)
    }

    pub async fn create_user(&self, new_user: &NewUser) -> ApiResult<String>
    {
        let request_body = RequestBody { data: new_user };

        let res = self.client.post(self.api_url.join("users")?)
            .json(&request_body)
            .send().await?;

        let body = res.text().await?;

        Self::get_api_result(&body)
    }

    pub async fn update_hidden_by_op(&self, updated_hidden_by_op: &UpdatedHiddenByOp, access_token: Option<&String>)
                                     -> ApiResult<String>
    {
        let request_body = RequestBody { data: updated_hidden_by_op };

        let res = Self::add_auth_header(
            self.client.patch(self.api_url.join("posts")?),
            access_token,
        ).json(&request_body).send().await?;

        let body = res.text().await?;

        Self::get_api_result(&body)
    }

    fn add_auth_header(req: reqwest::RequestBuilder, access_token: Option<&String>)
                       -> reqwest::RequestBuilder
    {
        if let Some(access_token) = access_token {
            return req.header("Authorization", access_token);
        }
        req
    }

    fn get_api_result<'a, T: de::Deserialize<'a>>(body: &'a str) -> ApiResult<T> {
        let res = match serde_json::from_str::<SuccessResponse<T>>(body) {
            Ok(sr) => Ok(sr),
            Err(_) => Err(serde_json::from_str::<ErrorResponse>(body)?),
        };
        Ok(res)
    }

    fn get_api_url(url: &String) -> Result<Url, Box<dyn std::error::Error>> {
        Ok(Url::parse(url)?.join("/s_api/")?)
    }

    fn build_client(settings: &ApiClientSettings) -> Result<Client, Box<dyn std::error::Error>> {
        let proxy = settings.proxy.as_ref().and_then(
            |proxy| if proxy == "" { None } else { Some(proxy) }
        );
        let proxy = match proxy {
            Some(proxy) => Some(reqwest::Proxy::all(proxy)?),
            None => None,
        };

        let client_builder = reqwest::Client::builder()
            .user_agent(Self::get_user_agent())
            .gzip(true)
            .timeout(Duration::from_secs(settings.request_timeout));

        let client_builder = match proxy {
            Some(proxy) => client_builder.proxy(proxy),
            None => client_builder,
        };

        Ok(client_builder.build()?)
    }

    fn get_user_agent() -> String {
        format!("pfpmd-ui3/{}", crate::constants::VERSION)
    }

    fn get_post_text_html(post: &Post, db: &sled::Db) -> Result<String, Box<dyn std::error::Error>> {
        let post_id = post.post_id.as_bytes();

        let stored_value = db.get(post_id).map_err(|err| {
            error!("Failed to get post {} from db: {}", post.post_id, err);
            err
        })?;

        Ok(match stored_value {
            None => {
                let value = Markup::parse_to_html(&post.text);
                db.insert(post_id, value.as_bytes()).map_err(|err| {
                    error!("Failed to insert post {} into db: {}", post.post_id, err);
                    err
                })?;
                value
            }
            Some(stored_value) => {
                let value = str::from_utf8(&stored_value).map_err(|err| {
                    error!("Failed to deserialize post {} from db: {}", post.post_id, err);
                    err
                })?;
                String::from(value)
            }
        })
    }
}