#![forbid(unsafe_code)]

#[macro_use]
extern crate lazy_static;
#[macro_use]
extern crate log;
#[macro_use]
extern crate serde_json;
#[macro_use]
extern crate handlebars;

mod models;
mod parsers;
mod settings;
mod validators;
mod routes;
mod api_client;
mod constants;
mod pages;
mod templates;
mod pagination;
mod serializers;
mod db_service;

use settings::Settings;
use std::net::SocketAddr;
use crate::routes::get_routes;
use crate::api_client::ApiClient;
use crate::templates::init_templates;
use handlebars::Handlebars;
use std::sync::atomic::{AtomicU32, Ordering};
use std::sync::Arc;

lazy_static! {
    pub static ref SETTINGS: Settings = Settings::new().unwrap();
    pub static ref API_CLIENT: ApiClient = ApiClient::new(&SETTINGS).unwrap();
    pub static ref HB: Handlebars<'static> = init_templates().unwrap();
    pub static ref POW_BITS: Arc<AtomicU32> = Arc::new(AtomicU32::new(100));
}

#[tokio::main]
async fn main() {
    pretty_env_logger::init();

    match API_CLIENT.connect().await {
        Err(err) => {
            error!("Failed to connect to the server: {}", err);
            return;
        },
        Ok(pow_bits) => POW_BITS.store(pow_bits, Ordering::Relaxed),
    };

    let addr: SocketAddr = (SETTINGS.server.host, SETTINGS.server.port).into();

    println!("Listening on http://{}", addr);

    warp::serve(get_routes()).run(addr).await;
}

fn get_static_ctx<'a>() -> (&'static Settings, &'static ApiClient, &'static Handlebars<'a>) {
    let settings: &Settings = &SETTINGS;
    let api_client: &ApiClient = &API_CLIENT;
    let hb: &Handlebars = &HB;

    (settings, api_client, hb)
}
