use serde_derive::{Deserialize, Serialize};
use std::error::Error as StdError;
use std::fmt::Display;
use core::fmt;
use uuid::Uuid;
use chrono::{DateTime, Utc};
use crate::serializers::*;

#[derive(Debug, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct SuccessResponse<T> {
    pub data: T,

    #[serde(flatten)]
    pub pagination: Option<Pagination>,

    pub meta: Option<serde_json::Value>,
}

#[derive(Debug, Deserialize)]
pub struct Pagination {
    pub skipped: u32,
    pub taken: u32,
    pub total: Option<i64>,
}

#[derive(Debug, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct ErrorResponse {
    pub error_code: ErrorCode,
    pub message: String,
}

#[derive(Debug, Deserialize)]
pub enum ErrorCode {
    E0400,
    E0401,
    E0403,
    E0404,
    E0411,
    E0413,
    E0415,
    E0500,
}

impl StdError for ErrorResponse {}

impl Display for ErrorResponse {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result { f.write_str("") }
}

#[derive(Debug, Serialize, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct Board {
    pub name: String,
    pub description: String,
}

#[derive(Debug, Serialize, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct Thread {
    #[serde(serialize_with = "to_simple_uuid")]
    pub thread_id: Uuid,
    pub subject: String,
    pub post_count: u32,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub board: Option<Board>,
    #[serde(skip_serializing)]
    pub last_post_date: Option<DateTime<Utc>>,
}

#[derive(Debug, Deserialize, Serialize)]
#[serde(rename_all = "camelCase")]
pub struct ShortUserInfo {
    pub nickname: String,
    pub banned: bool,
}

#[derive(Debug, Serialize, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct Post {
    #[serde(serialize_with = "to_simple_uuid")]
    pub post_id: Uuid,
    #[serde(serialize_with = "to_date_time")]
    pub post_date: DateTime<Utc>,
    #[serde(serialize_with = "to_hsl")]
    pub bg: u32,
    #[serde(serialize_with = "to_hsl")]
    pub fg: u32,
    #[serde(rename(serialize = "textHtml"))]
    pub text: String,
    pub hidden_by_op: bool,
    pub user: ShortUserInfo,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub thread: Option<Thread>,
    #[serde(skip_serializing)]
    pub index: Option<i64>,
}

#[derive(Debug, Serialize, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct ThreadSlice {
    pub thread: Thread,
    pub op: Post,
    pub posts: Vec<Post>,
    #[serde(skip_deserializing, skip_serializing_if = "Option::is_none")]
    pub omitted_count: Option<u32>,
}

#[derive(Debug, Deserialize, Serialize)]
#[serde(rename_all = "camelCase")]
pub struct WarpSlice {
    #[serde(skip_serializing_if = "Option::is_none")]
    pub board: Option<Board>,
    pub shell_threads: Vec<ThreadSlice>,
}

#[derive(Debug, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct ServerInfo {
    pub version: String,
    pub pow_bits: u32,
}

#[derive(Debug, Serialize)]
#[serde(rename_all = "camelCase")]
#[serde(tag = "type")]
pub enum PaginationElement {
    Gap {},

    #[serde(rename_all = "camelCase")]
    Link { page: u32, href: String, is_selected: bool },
}

/// This struct is used to create a new post/thread/board.
#[derive(Debug, Serialize, Deserialize, Clone)]
#[serde(rename_all = "camelCase")]
pub struct NewPost {
    #[serde(default)]
    pub client_id: String,
    pub post_id: Option<Uuid>,
    pub text: String,
    pub thread_id: Option<Uuid>, // if such thread doesn't exist, create a new one
    pub subject: Option<String>,
    pub board_name: Option<String>, // if such board doesn't exist, create a new one

    #[serde(default)]
    pub stay_here: bool,
    pub back_url: Option<String>,
}

#[derive(Debug, Deserialize, Serialize)]
#[serde(rename_all = "camelCase")]
pub struct UpdatedHiddenByOp {
    #[serde(skip_deserializing)]
    pub client_id: String,
    pub post_id: Uuid,
    pub hidden_by_op: bool,
    #[serde(skip_serializing)]
    pub back_url: String,
}

#[derive(Debug, Serialize)]
#[serde(rename_all = "camelCase")]
pub struct RequestBody<T> {
    pub data: T,
}

#[derive(Debug, Serialize, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct Claims {
    pub name: String,
    pub iat: i64,
}

#[derive(Debug, Serialize, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct Cookies {
    pub client_id: Option<String>,

    #[serde(skip_serializing)]
    pub access_token: Option<String>,

    pub fix_menu: bool,
    pub show_thumbs: bool,
    pub max_width: u32,
    pub compact: bool,
    pub only_chat_posts: bool,
}

#[derive(Debug, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct UserSettings {
    #[serde(default)]
    pub change_colors: bool,
    #[serde(default)]
    pub fix_menu: bool,
    #[serde(default)]
    pub show_thumbs: bool,
    pub max_width: u32,
    #[serde(default)]
    pub compact_theme: bool,
    #[serde(default)]
    pub only_chat_posts: bool,
    pub client_id: Option<String>,
    pub back_url: Option<String>,
}

#[derive(Debug, Serialize, Deserialize, Clone)]
#[serde(rename_all = "camelCase")]
pub struct Auth {
    pub nickname: String,
    pub password: String,

    #[serde(skip_serializing_if = "Option::is_none")]
    pub back_url: Option<String>,
}

#[derive(Debug, Deserialize, Serialize, Clone)]
#[serde(rename_all = "camelCase")]
pub struct NewUser {
    pub nickname: String,
    pub password: String,
    pub pow: String,
}

#[derive(Debug, Deserialize, Serialize)]
#[serde(rename_all = "camelCase")]
pub struct AuthInfo {
    pub access_token: String,
}

#[derive(Debug, Serialize, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct UiState {
    pub cookies: Cookies,
    pub ui: String,
    pub version: String,
}
