use crate::models::*;
use crate::pagination::{get_page_number, get_pagination_elements, get_thread_page_number};
use uuid::Uuid;
use warp::Reply;
use warp::multipart::{FormData, Part};
use warp::http;
use warp::http::header::{HeaderName, HeaderValue, CONTENT_TYPE};
use std::convert::TryFrom;
use std::sync::Arc;
use warp::reply::Response;
use std::ops::Deref;
use crate::constants::re;
use crate::parsers::{zip_option, zip_option3};
use regex::Regex;
use serde::Serialize;
use serde_json;
use futures::TryStreamExt;
use bytes::BufMut;
use std::sync::atomic::Ordering;

pub async fn boards_page(page: Option<u32>, cookies: Cookies, req_et: Option<String>)
                         -> Result<Response, Response>
{
    let (settings, api_client, hb) = crate::get_static_ctx();

    let page_size = settings.ui.page_size;
    let skip = page.unwrap_or(0) * page_size;

    let SuccessResponse {
        data: boards,
        pagination,
        meta: _,
    } = api_client.get_boards(skip, page_size).await.map_err(map_api_err)?;

    let pagination = get_pagination(pagination)?;

    let pagination_elements = get_pagination_elements(
        pagination.get_total()?,
        get_page_number(pagination.skipped, page_size),
        page_size,
        settings.ui.pagination_sibling_count,
        |page| if page == 0 {
            format!("{}/boards", settings.ui.base_path)
        } else {
            format!("{}/boards?p={}", settings.ui.base_path, page)
        },
    );

    let data = json!({
        "pageName": "boards",
        "title": "Boards",
        "basePath": settings.ui.base_path,
        "boards": &boards,
        "paginationElements": pagination_elements,
        "cookies": &cookies,
        "claims": get_claims(&cookies).ok(),
    });

    reply_html(
        hb.render("boards", &data).map_err(map_render_err)?,
        req_et.as_deref(),
        None,
        cookies.client_id.is_none(),
        None::<fn() -> _>,
    )
}

pub async fn thread_page(thread_id: Uuid, page: Option<u32>, cookies: Cookies,
                         req_et: Option<String>, db: Arc<sled::Db>)
                         -> Result<Response, Response>
{
    let (settings, api_client, hb) = crate::get_static_ctx();

    let page_size = settings.ui.page_size;
    let skip = page.unwrap_or(0) * page_size;

    let SuccessResponse {
        data: thread_slice,
        pagination,
        meta: _,
    } = api_client.get_thread_slice(
        &thread_id,
        skip,
        page_size,
        page.is_none(),
        db.deref(),
    ).await.map_err(map_api_err)?;

    let pagination = get_pagination(pagination)?;
    let board = thread_slice.get_board()?;

    let current_page = get_thread_page_number(pagination.skipped, page_size);

    let thread_id_simple = thread_id.to_simple();

    let pagination_elements = get_pagination_elements(
        pagination.get_total()?,
        current_page,
        page_size,
        settings.ui.pagination_sibling_count,
        |page|
            format!("{}/{}/{}?p={}", settings.ui.base_path, board.name, &thread_id_simple, page),
    );

    let title = format!("/{}/{} – ``{}'' – page {}",
                        board.name, &thread_id_simple.to_string()[0..5],
                        thread_slice.thread.subject, current_page);

    let data = json!({
        "pageName": "thread",
        "title": title,
        "basePath": settings.ui.base_path,
        "threadSlice": &thread_slice,
        "paginationElements": &pagination_elements,
        "cookies": &cookies,
        "claims": get_claims(&cookies).ok(),
    });

    reply_html(
        hb.render("thread", &data).map_err(map_render_err)?,
        req_et.as_deref(),
//        if pagination.taken == page_size { Some(60 * 60) } else { None },
        None,
        cookies.client_id.is_none(),
        None::<fn() -> _>,
    )
}

pub async fn warp_page(board_name: Option<String>, page: Option<u32>, cookies: Cookies,
                       req_et: Option<String>, db: Arc<sled::Db>)
                       -> Result<Response, Response>
{
    let (settings, api_client, hb) = crate::get_static_ctx();

    let page_size = settings.ui.warp_page_size;
    let skip = page.unwrap_or(0) * page_size;

    let SuccessResponse {
        data: warp_slice,
        pagination,
        meta: _,
    } = api_client.get_warp_slice(board_name.as_ref(), skip, page_size, db.deref())
        .await
        .map_err(map_api_err)?;

    let pagination = get_pagination(pagination)?;

    let board = board_name.as_ref().and_then(|_| warp_slice.board.as_ref());

    let current_page = get_page_number(pagination.skipped, page_size);

    let pagination_elements = get_pagination_elements(
        pagination.get_total()?,
        current_page,
        page_size,
        settings.ui.pagination_sibling_count,
        |page|
            match board_name.as_ref() {
                None => format!("{}/?p={}", settings.ui.base_path, page),
                Some(board_name) => format!("{}/{}?p={}", settings.ui.base_path, board_name, page)
            },
    );

    let title = match board {
        None => if current_page == 0 {
            "pfpmd".to_string()
        } else {
            format!("pfpmd – page {}", current_page)
        },
        Some(board) => {
            let board_description_part = if board.description.is_empty() {
                "".to_string()
            } else {
                format!(" – ``{}''", board.description)
            };

            if current_page == 0 {
                format!("/{}/{}", board.name, board_description_part)
            } else {
                format!("/{}/{} – page {}", board.name, board_description_part, current_page)
            }
        }
    };

    let data = json!({
        "pageName": if board.is_none() { "warp" } else { "board" },
        "title": title,
        "basePath": settings.ui.base_path,
        "warpSlice": &warp_slice,
        "paginationElements": &pagination_elements,
        "cookies": &cookies,
        "claims": get_claims(&cookies).ok(),
    });

    reply_html(
        hb.render("warp", &data).map_err(map_render_err)?,
        req_et.as_deref(),
        None,
        cookies.client_id.is_none(),
        None::<fn() -> _>,
    )
}

pub async fn post_page(post_id: Uuid) -> Result<Response, Response>
{
    let (settings, api_client, _) = crate::get_static_ctx();

    let page_size = settings.ui.page_size;

    let SuccessResponse {
        data: post,
        pagination: _,
        meta: _,
    } = api_client.get_post(&post_id).await.map_err(map_api_err)?;

    let skipped = post.get_index()?;

    let current_page = get_thread_page_number(skipped, page_size);

    let thread = post.get_thread()?;
    let board = post.get_board()?;

    let redirect_url = format!(
        "{}/{}/{}?p={}#{}",
        settings.ui.base_path, board.name,
        thread.thread_id.to_simple(), current_page, post.post_id.to_simple()
    );

    Ok(warp::reply::with_header(
        http::StatusCode::MOVED_PERMANENTLY,
        http::header::LOCATION,
        redirect_url,
    ).into_response())
}

pub async fn chat_page(cookies: Cookies, req_et: Option<String>) -> Result<Response, Response> {
    let (settings, _, hb) = crate::get_static_ctx();

    let data = json!({
        "pageName": "chat",
        "title": "Chat",
        "basePath": settings.ui.base_path,
        "chatThreadId": crate::constants::CHAT_THREAD_ID,
        "cookies": &cookies,
        "claims": get_claims(&cookies).ok(),
    });

    reply_html(
        hb.render("chat", &data).map_err(map_render_err)?,
        req_et.as_deref(),
        None,
        cookies.client_id.is_none(),
        None::<fn() -> _>,
    )
}

pub async fn chat_frame(cookies: Cookies, req_et: Option<String>, db: Arc<sled::Db>)
                        -> Result<Response, Response>
{
    let (settings, api_client, hb) = crate::get_static_ctx();

    let SuccessResponse {
        data: posts,
        pagination: _,
        meta: _,
    } = api_client.get_recent_posts(
        if cookies.only_chat_posts { Some("cht") } else { None },
        db.deref(),
    )
        .await
        .map_err(map_api_err)?;

    let data = json!({
        "basePath": settings.ui.base_path,
        "posts": &posts,
        "cookies": &cookies,
        "claims": get_claims(&cookies).ok(),
    });

    reply_html(
        hb.render("chat_frame", &data).map_err(map_render_err)?,
        req_et.as_deref(),
        None,
        cookies.client_id.is_none(),
        None::<fn() -> _>,
    )
}

pub async fn help_page(mut cookies: Cookies, req_et: Option<String>)
                       -> Result<Response, Response>
{
    let (settings, _, hb) = crate::get_static_ctx();

    let was_client_set = cookies.client_id.is_some();

    cookies.client_id = Some(cookies.client_id.unwrap_or_else(|| new_client_id()));

    let data = json!({
        "pageName": "help",
        "title": "Help",
        "basePath": settings.ui.base_path,
        "cookies": &cookies,
        "claims": get_claims(&cookies).ok(),
    });

    reply_html(
        hb.render("help", &data).map_err(map_render_err)?,
        req_et.as_deref(),
        None,
        !was_client_set,
        Some(move || cookies.client_id.unwrap()),
    )
}

pub async fn settings_page(cookies: Cookies, req_et: Option<String>, referer: Option<String>)
                           -> Result<Response, Response>
{
    let (settings, _, hb) = crate::get_static_ctx();

    let data = json!({
        "pageName": "settings",
        "title": "Settings",
        "basePath": settings.ui.base_path,
        "cookies": &cookies,
        "claims": get_claims(&cookies).ok(),
        "backUrl": referer,
    });

    reply_html(
        hb.render("settings", &data).map_err(map_render_err)?,
        req_et.as_deref(),
        None,
        cookies.client_id.is_none(),
        None::<fn() -> _>,
    )
}

pub async fn login_page(cookies: Cookies, req_et: Option<String>, referer: Option<String>)
                        -> Result<Response, Response>
{
    let (settings, _, hb) = crate::get_static_ctx();

    let data = json!({
        "pageName": "login",
        "title": "Login / Create Account",
        "basePath": settings.ui.base_path,
        "cookies": &cookies,
        "claims": get_claims(&cookies).ok(),
        "powBits": crate::POW_BITS.load(Ordering::Relaxed),
        "backUrl": referer,
    });

    reply_html(
        hb.render("login", &data).map_err(map_render_err)?,
        req_et.as_deref(),
        None,
        cookies.client_id.is_none(),
        None::<fn() -> _>,
    )
}

pub async fn profile_page(cookies: Cookies, req_et: Option<String>, referer: Option<String>)
                        -> Result<Response, Response>
{
    let (settings, _, hb) = crate::get_static_ctx();

    let data = json!({
        "pageName": "profile",
        "title": "Profile",
        "basePath": settings.ui.base_path,
        "cookies": &cookies,
        "claims": get_claims(&cookies).ok(),
        "backUrl": referer,
    });

    reply_html(
        hb.render("profile", &data).map_err(map_render_err)?,
        req_et.as_deref(),
        None,
        cookies.client_id.is_none(),
        None::<fn() -> _>,
    )
}

pub async fn import_page(cookies: Cookies, req_et: Option<String>, referer: Option<String>)
                         -> Result<Response, Response>
{
    let (settings, _, hb) = crate::get_static_ctx();

    let data = json!({
        "pageName": "import",
        "title": "Import",
        "basePath": settings.ui.base_path,
        "cookies": &cookies,
        "claims": get_claims(&cookies).ok(),
        "backUrl": referer,
    });

    reply_html(
        hb.render("import", &data).map_err(map_render_err)?,
        req_et.as_deref(),
        None,
        cookies.client_id.is_none(),
        None::<fn() -> _>,
    )
}

pub async fn export(cookies: Cookies, req_et: Option<String>)
                    -> Result<Response, Response>
{
    let was_client_set = cookies.client_id.is_some();

    let data = UiState {
        ui: "ui3".to_string(),
        version: crate::constants::VERSION.to_string(),
        cookies,
    };

    reply_json(
        data,
        req_et.as_deref(),
        None,
        !was_client_set,
        None::<fn() -> _>,
    )
}

pub async fn publish(mut new_post: NewPost, cookies: Cookies, referer: Option<String>)
                     -> Result<Response, Response>
{
    let (settings, api_client, _) = crate::get_static_ctx();

    let redirect_url = new_post.back_url.to_owned().or(referer)
        .unwrap_or(format!("{}/", settings.ui.base_path));

    // check commands first
    let mut text = new_post.text.trim();

    if text == "/help" || text == "/settings" {
        return Ok(found(&format!("{}{}", settings.ui.base_path, text)));
    }
    if text == "/cc" {
        let mut response = found(&redirect_url);
        let client_id = new_client_id();

        set_cookie(&mut response, format!("clientId={}", client_id))?;
        return Ok(response);
    }

    let mut command_handled = false;

    {
        // e.g. "/b random"
        lazy_static! {
            static ref RE: Regex = Regex::new(&format!("^/({}) (.*)$", re::BOARD)).unwrap();
        }

        let caps: Option<regex::Captures> = RE.captures(text);

        if let Some((board_name, description)) = caps.and_then(
            |caps| zip_option(caps.get(1), caps.get(2))
        ) {
            let board = Board {
                name: board_name.as_str().to_string(),
                description: description.as_str().to_string(),
            };
            let res = api_client.update_board(board, cookies.access_token.as_ref()).await
                .map_err(map_api_err)?;

            return match res {
                Err(e) => reply_html_error(&e.message, &cookies),
                Ok(_) => Ok(found(&redirect_url))
            };
        }
    }

    {
        // e.g. "/b/7f2176daf34f494583c273bebceb22a1 post text"
        lazy_static! {
            static ref RE: Regex = Regex::new(
                &format!("^/({})/([0-9A-Za-z\\-]+) (.*)$", re::BOARD)
            ).unwrap();
        }

        let caps: Option<regex::Captures> = RE.captures(text);

        if let Some((board_name, thread_id, post_text)) = caps.and_then(
            |caps| zip_option3(caps.get(1), caps.get(2), caps.get(3))
        ) {
            if let Ok(thread_id) = Uuid::parse_str(thread_id.as_str()) {
                new_post.board_name = Some(board_name.as_str().to_string());
                new_post.thread_id = Some(thread_id);
                new_post.subject = None;
                new_post.text = post_text.as_str().to_string();

                text = &new_post.text;
                command_handled = true;
            }
        }
    }

    if !command_handled {
        // e.g. "/b[subject] post text"
        lazy_static! {
            static ref RE: Regex = Regex::new(&format!("^/({})\\[(.+)\\] (.*)$", re::BOARD)).unwrap();
        }

        let caps: Option<regex::Captures> = RE.captures(text);

        if let Some((board_name, subject, post_text)) = caps.and_then(
            |caps| zip_option3(caps.get(1), caps.get(2), caps.get(3))
        ) {
            new_post.board_name = Some(board_name.as_str().to_string());
            new_post.thread_id = None;
            new_post.subject = Some(subject.as_str().to_string());
            new_post.text = post_text.as_str().to_string();
        }
    }

    new_post.client_id = cookies.client_id.as_ref().ok_or_else(|| {
        error!("Publish failed because `client_id` is None");
        found(&redirect_url)
    })?.clone();

    let stay_here = new_post.stay_here;

    let res = api_client.create_post(new_post, cookies.access_token.as_ref()).await
        .map_err(map_api_err)?;

    match res {
        Err(e) => reply_html_error(&e.message, &cookies),
        Ok(sr) => {
            let post = sr.data;

            let thread = post.get_thread()?;
            let board = post.get_board()?;

            let redirect_url = if stay_here {
                redirect_url
            } else {
                format!(
                    "{}/{}/{}#{}", settings.ui.base_path, board.name,
                    thread.thread_id.to_simple(), post.post_id.to_simple()
                )
            };

            Ok(found(&redirect_url))
        }
    }
}

pub async fn update_hidden_by_op(mut updated_hidden_by_op: UpdatedHiddenByOp, cookies: Cookies,
                                 referer: Option<String>)
                                 -> Result<Response, Response>
{
    updated_hidden_by_op.client_id = cookies.client_id.as_ref().unwrap_or(&"".to_string()).to_owned();

    let (_settings, api_client, _hb) = crate::get_static_ctx();

    let res = api_client.update_hidden_by_op(
        &updated_hidden_by_op, cookies.access_token.as_ref(),
    ).await.map_err(map_api_err)?;

    match res {
        Err(e) => if e.message == "Invalid nickname" || e.message == "Invalid clientId" {
            reply_html_error(&format!("You're not OP of this thread: {}", e.message), &cookies)
        } else {
            reply_html_error(&e.message, &cookies)
        },
        Ok(_) => Ok(found(referer.as_ref().unwrap_or(&updated_hidden_by_op.back_url))),
    }
}

pub async fn save_settings(user_settings: UserSettings, cookies: Cookies)
                           -> Result<Response, Response>
{
    let (settings, _, _) = crate::get_static_ctx();

    let redirect_url = &user_settings.back_url.filter(|back_url|
        !back_url.is_empty()
            && !back_url.contains("settings")
            && !back_url.contains("saveSettings")
            && !back_url.contains("authorize")
            && !back_url.contains("publish")
            && !back_url.contains("signup")
    ).unwrap_or(
        format!("{}/", settings.ui.base_path)
    );

    let client_id = match (cookies.client_id, user_settings.change_colors, user_settings.client_id) {
        (_, false, Some(client_id)) => client_id,
        (Some(client_id), false, _) => client_id,
        _ => new_client_id(),
    };

    let mut response = found(redirect_url);

    set_cookie(&mut response, format!("clientId={}", client_id))?;
    set_cookie(&mut response, format!("fixMenu={}", user_settings.fix_menu))?;
    set_cookie(&mut response, format!("showThumbs={}", user_settings.show_thumbs))?;
    set_cookie(&mut response, format!("maxWidth={}", user_settings.max_width))?;
    set_cookie(&mut response, format!("compact={}", user_settings.compact_theme))?;
    set_cookie(&mut response, format!("onlyChatPosts={}", user_settings.only_chat_posts))?;

    Ok(response)
}

pub async fn authorize(auth: Auth, cookies: Cookies) -> Result<Response, Response>
{
    let (settings, api_client, _) = crate::get_static_ctx();

    let res = api_client.authorize(&auth)
        .await
        .map_err(map_api_err)?;

    match res {
        Err(e) => reply_html_error(&e.message, &cookies),
        Ok(sr) => {
            let redirect_url = &auth.back_url.filter(|back_url|
                !back_url.is_empty()
                    && !back_url.contains("login")
                    && !back_url.contains("authorize")
                    && !back_url.contains("publish")
                    && !back_url.contains("signup")
            ).unwrap_or(
                format!("{}/", settings.ui.base_path)
            );

            let client_id = match cookies.client_id {
                Some(client_id) => client_id,
                _ => new_client_id(),
            };

            let mut response = found(redirect_url);

            set_cookie(&mut response, format!("clientId={}", client_id))?;
            set_cookie(&mut response, format!("accessToken={}; path=/; samesite=strict", sr.data.access_token))?;

            Ok(response)
        }
    }
}

pub async fn logout(cookies: Cookies) -> Result<Response, Response>
{
    let (settings, _, _) = crate::get_static_ctx();

    let redirect_url = format!("{}/", settings.ui.base_path);

    let client_id = match cookies.client_id {
        Some(client_id) => client_id,
        _ => new_client_id(),
    };

    let mut response = found(&redirect_url);

    set_cookie(&mut response, format!("clientId={}", client_id))?;
    set_cookie(&mut response, "accessToken=; max-age=-1; path=/; samesite=strict".to_string())?;

    Ok(response)
}

pub async fn signup(new_user: NewUser, cookies: Cookies) -> Result<Response, Response>
{
    let (settings, api_client, hb) = crate::get_static_ctx();

    let res = api_client.create_user(&new_user)
        .await
        .map_err(map_api_err)?;

    match res {
        Err(e) => reply_html_error(&e.message, &cookies),
        Ok(_) => {
            let message = "User created. Now you can log in.";

            let data = json!({
                "basePath": settings.ui.base_path,
                "title": "Success",
                "cookies": &cookies,
                "claims": get_claims(&cookies).ok(),
                "message": message,
            });

            reply_html(
                hb.render("success", &data).map_err(map_render_err)?,
                None,
                None,
                cookies.client_id.is_none(),
                None::<fn() -> _>,
            )
        }
    }
}

pub async fn import_ui_state(form: FormData, _cookies: Cookies)
                             -> Result<Response, Response>
{
    let parts: Vec<Part> = form.try_collect().await.map_err(|err| {
        error!("multipart form error: {}", err);
        error::internal("form data error")
    })?;

    let mut ui_state: Option<UiState> = None;
    let mut back_url: Option<String> = None;

    for p in parts {
        if p.name() == "uiState" {
            if p.content_type() != Some("application/json") {
                error!("invalid content-type of uiState");
                return Err(error::internal("invalid content-type"));
            }

            let value = get_part_value(p).await?;

            ui_state = Some({
                let ui_state: UiState = serde_json::from_slice(&value).map_err(|err| {
                    error!("Failed to parse uiState: {}", err);
                    error::internal(&format!("Failed to parse uiState: {}", err))
                })?;

                if &ui_state.ui != "ui3" {
                    return Err(error::internal(
                        &format!("Invalid ui: expected `ui3`, got `{}`", &ui_state.ui)
                    ));
                }
                ui_state
            });
        } else if p.name() == "backUrl" {
            let value = get_part_value(p).await?;
            back_url = Some(std::str::from_utf8(&value).map_err(|err| {
                error!("Failed to parse backUrl: {}", err);
                error::internal("Failed to parse backUrl")
            })?.to_string());
        }
    }

    let ui_state = ui_state.ok_or_else(|| {
        error!("uiState is required");
        error::internal("uiState is required")
    })?;

    let (settings, _, _) = crate::get_static_ctx();

    let redirect_url = back_url.filter(|back_url|
        !back_url.is_empty() && !back_url.contains("import")
    ).unwrap_or(
        format!("{}/", settings.ui.base_path)
    );

    let mut response = found(&redirect_url);

    let c = &ui_state.cookies;

    if let Some(client_id) = c.client_id.as_ref() {
        set_cookie(&mut response, format!("clientId={}", client_id))?;
    }
    set_cookie(&mut response, format!("fixMenu={}", c.fix_menu))?;
    set_cookie(&mut response, format!("showThumbs={}", c.show_thumbs))?;
    set_cookie(&mut response, format!("maxWidth={}", c.max_width))?;
    set_cookie(&mut response, format!("compact={}", c.compact))?;
    set_cookie(&mut response, format!("onlyChatPosts={}", c.only_chat_posts))?;

    Ok(response)
}

fn new_client_id() -> String {
    use rand::Rng;
    use rand::distributions::Alphanumeric;

    rand::thread_rng()
        .sample_iter(&Alphanumeric)
        .take(10)
        .collect::<String>()
}

fn get_claims(cookies: &Cookies) -> Result<Claims, String> {
    let error_invalid_token = "Invalid access token";

    let access_token = cookies.access_token.as_ref().ok_or(error_invalid_token.to_string())?;

    let parts: Vec<&str> = access_token.split('.').collect();

    if parts.len() != 3 {
        return Err(error_invalid_token.to_string());
    }
    let claims_str = parts[1];
    let claims_bytes = base64::decode(claims_str)
        .map_err(|_| error_invalid_token.to_string())?;

    let claims = serde_json::from_slice::<Claims>(&claims_bytes)
        .map_err(|_| error_invalid_token.to_string())?;

    Ok(claims)
}

fn found(redirect_url: &str) -> Response {
    warp::reply::with_header(
        http::StatusCode::FOUND,
        http::header::LOCATION,
        redirect_url,
    ).into_response()
}

fn reply_html_error(message: &str, cookies: &Cookies) -> Result<Response, Response> {
    let (settings, _, hb) = crate::get_static_ctx();

    let message = if message == "Unauthorized" {
        "Unauthorized. Please log in."
    } else {
        message
    };

    let data = json!({
        "basePath": settings.ui.base_path,
        "title": "Error",
        "cookies": &cookies,
        "claims": get_claims(&cookies).ok(),
        "message": message,
    });

    reply_html(
        hb.render("error", &data).map_err(map_render_err)?,
        None,
        None,
        cookies.client_id.is_none(),
        None::<fn() -> _>,
    )
}

/// NOTE: use it only for GET requests
fn reply_html(html: String, request_etag: Option<&str>, max_age: Option<u32>, set_client_id: bool,
              get_client_id: Option<impl FnOnce() -> String>)
              -> Result<Response, Response>
{
    reply(ResponseBody::<String>::Html(html), request_etag, max_age, set_client_id, get_client_id)
}

fn reply_json<T>(json: T, request_etag: Option<&str>, max_age: Option<u32>, set_client_id: bool,
                 get_client_id: Option<impl FnOnce() -> String>)
                 -> Result<Response, Response>
    where T: Serialize
{
    reply(ResponseBody::Json(json), request_etag, max_age, set_client_id, get_client_id)
}

fn reply<T>(body: ResponseBody<T>, request_etag: Option<&str>, max_age: Option<u32>, set_client_id: bool,
            get_client_id: Option<impl FnOnce() -> String>)
            -> Result<Response, Response>
    where T: Serialize
{
    let json_bytes: Option<Vec<u8>> = match &body {
        ResponseBody::Json(json) => serde_json::to_vec_pretty(json).ok(),
        _ => None,
    };

    let mut res_et = etag::EntityTag::from_hash(match &body {
        ResponseBody::Html(html) => html.as_bytes(),
        ResponseBody::Json(_) => json_bytes.as_ref().unwrap(),
    });
    res_et.weak = true;
    let res_et = res_et.to_string();

    let mut response = request_etag.and_then(|req_et| {
        if req_et == &res_et {
            http::Response::builder()
                .status(http::status::StatusCode::NOT_MODIFIED)
                .body(warp::hyper::Body::empty())
                .map_err(|err| {
                    error!("Failed to build `304 Not Modified` response: {}", err);
                    error::internal("failed to build `304 Not Modified` response")
                })
                .ok()
        } else {
            None
        }
    }).unwrap_or_else(|| match body {
        ResponseBody::Html(html) => warp::reply::html(html).into_response(),
        ResponseBody::Json(_) => {
            let mut res = Response::new(json_bytes.unwrap().into());
            res.headers_mut()
                .insert(CONTENT_TYPE, HeaderValue::from_static("application/json"));
            res
        }
    });

    set_header(&mut response, http::header::ETAG, res_et)?;

    if set_client_id {
        let client_id = get_client_id
            .map(|get_client_id| get_client_id())
            .unwrap_or_else(|| new_client_id());

        set_cookie(&mut response, format!("clientId={}", client_id))?;
    }

    if let Some(max_age) = max_age {
        set_header(
            &mut response,
            http::header::CACHE_CONTROL,
            format!("public, max-age={}", max_age),
        )?;
    }

    Ok(response)
}

enum ResponseBody<T: Serialize> {
    Html(String),
    Json(T),
}

fn set_header(response: &mut http::Response<warp::hyper::Body>, name: HeaderName, value: String)
              -> Result<(), Response>
{
    HeaderValue::try_from(&value).map(|value| {
        response.headers_mut().append(&name, value);
    }).map_err(|_err| {
        error!("Failed to add a response header {} = {}", name, value);
        error::internal("failed to set a response header")
    })
}

fn set_cookie(response: &mut http::Response<warp::hyper::Body>, value: String)
              -> Result<(), Response>
{
    set_header(response, http::header::SET_COOKIE, value)
}

fn map_api_err(err: Box<dyn std::error::Error>) -> Response {
    error!("Api request failed: {:?}", err);
    error::internal("api request has failed")
}

fn get_pagination(pagination: Option<Pagination>) -> Result<Pagination, Response> {
    pagination.ok_or_else(|| {
        error!("Api response doesn't have `pagination` but it was expected");
        error::internal("`pagination` data expected but not found in api response")
    })
}

impl Pagination {
    pub fn get_total(&self) -> Result<u32, Response> {
        self.total.and_then(|total| {
            u32::try_from(total).map_err(|err| {
                error!("Failed to convert `total` to u32: {}", err);
                error::internal("`total` from the api response is not valid")
            }).ok()
        }).ok_or_else(|| {
            error!("Api response doesn't have `total` but it was expected");
            error::internal("`total` expected but not found in api response")
        })
    }
}

impl Post {
    pub fn get_index(&self) -> Result<u32, Response> {
        self.index.and_then(|index| {
            u32::try_from(index).map_err(|err| {
                error!("Failed to convert a post's `index` to u32: {}", err);
                error::internal("`index` from the api response is not valid")
            }).ok()
        }).ok_or_else(|| {
            error!("Api response doesn't have `index` but it was expected");
            error::internal("`index` expected but not found in api response")
        })
    }

    pub fn get_thread(&self) -> Result<&Thread, Response> {
        self.thread.as_ref().ok_or_else(|| {
            error!("Api response fail: `post.thread` must not be None");
            error::internal("`thread` expected but not found in api response")
        })
    }

    pub fn get_board(&self) -> Result<&Board, Response> {
        self.get_thread()?.board.as_ref().ok_or_else(|| {
            error!("Api response fail: `post.thread.board` must not be None");
            error::internal("`board` expected but not found in api response")
        })
    }
}

impl ThreadSlice {
    pub fn get_board(&self) -> Result<&Board, Response> {
        self.thread.board.as_ref().ok_or_else(|| {
            error!("Api response fail: thread_slice.thread.board must not be None");
            error::internal("`board` expected but not found in api response")
        })
    }
}

fn map_render_err(err: handlebars::RenderError) -> Response {
    error!("Render fail: {}", err);
    error::internal("failed to render an html template")
}

mod error {
    use super::*;

    pub fn internal(body: &str) -> warp::reply::Response {
        warp::reply::with_status(
            format!("Internal server error: {}", body),
            http::status::StatusCode::INTERNAL_SERVER_ERROR,
        ).into_response()
    }
}

async fn get_part_value(part: Part) -> Result<Vec<u8>, Response> {
    let value = part
        .stream()
        .try_fold(Vec::new(), |mut vec, data| {
            vec.put(data);
            async move { Ok(vec) }
        })
        .await
        .map_err(|err| {
            error!("Failed to read the value of a part of the muplipart form: {}", err);
            error::internal("Failed to read the value of a part of the muplipart form")
        })?;

    Ok(value)
}
