use crate::models::PaginationElement;

pub fn get_thread_page_number(skipped: u32, page_size: u32) -> u32 {
    // Note: page number starts form 0
    // Note: skipped starts from 1 because we always skip OP

    get_page_number(skipped.saturating_sub(1), page_size)
}

pub fn get_page_number(skipped: u32, page_size: u32) -> u32 {
    // Note: page number starts form 0

    if skipped == 0 { 0 } else {
        (skipped as f64 / page_size as f64).floor() as u32
    }
}

pub fn get_pagination_elements<F>(
    total: u32,
    current_page: u32,
    page_size: u32,
    sibling_count: u32,
    get_href: F,
) -> Vec<PaginationElement>
    where F: Fn(u32) -> String
{
    assert_ne!(page_size, 0);

    let page_count = (total as f64 / page_size as f64).ceil() as u32;
    if page_count <= 1 { return vec![]; }

    let mut elements: Vec<PaginationElement> = vec![];

    let last_page = page_count - 1;

    if current_page > 0 {
        elements.push(PaginationElement::Link {
            page: 0,
            href: get_href(0),
            is_selected: false,
        });
    }

    // 0 . 2 3 4 [5]
    let is_left_gap_needed = current_page
        .saturating_sub(sibling_count)
        .saturating_sub(1) > 0;

    if is_left_gap_needed { elements.push(PaginationElement::Gap {}); }

    for i in (0..sibling_count).rev() {
        let page = current_page
            .saturating_sub(i)
            .saturating_sub(1);

        if page > 0 {
            elements.push(PaginationElement::Link {
                page,
                href: get_href(page),
                is_selected: false,
            });
        }
    }

    elements.push(PaginationElement::Link {
        page: current_page,
        href: get_href(current_page),
        is_selected: true,
    });

    for i in 0..sibling_count {
        let page = current_page + i + 1;

        if page < last_page {
            elements.push(PaginationElement::Link {
                page,
                href: get_href(page),
                is_selected: false,
            });
        }
    }

    // [5] 6 7 8 . 10
    let is_right_gap_needed = current_page + sibling_count + 1 < last_page;
    if is_right_gap_needed { elements.push(PaginationElement::Gap {}); }

    if current_page < last_page {
        elements.push(PaginationElement::Link {
            page: last_page,
            href: get_href(last_page),
            is_selected: false,
        });
    }

    elements
}