use nom::character::complete::{anychar, newline, multispace0, multispace1};
use nom::error::{ErrorKind, ParseError};
use nom::IResult;
use nom::sequence::{delimited, preceded, terminated, tuple};
use nom::bytes::complete::{tag, take_until, take_till, take_while_m_n};
use nom::branch::alt;
use htmlize::{escape_attribute, escape_text};
use nom::multi::many0;
use uuid::Uuid;
use regex::Regex;
use percent_encoding::{utf8_percent_encode, AsciiSet, CONTROLS, percent_decode_str};
use crate::settings::Settings;
use std::borrow::Cow;

#[derive(PartialEq, Debug)]
pub struct Code<'a> {
    pub lang: &'a str,
    pub code: &'a str,
}

#[derive(PartialEq, Debug)]
pub struct Link<'a> {
    pub name: Option<&'a str>,
    pub url: &'a str,
    pub media: Option<Media>,
}

#[derive(PartialEq, Debug)]
pub enum Media {
    Thumb(Thumb),
    #[allow(dead_code)]
    Embed(Embed),
}

#[derive(PartialEq, Debug)]
pub enum ThumbOf {
    Youtube,
    Imgur,
    Twitter,
    Ploudfr,
    LibreVideo,
    PeertubeSocial,
}

#[derive(PartialEq, Debug)]
pub enum EmbedOf {
    #[allow(dead_code)]
    Bitchute,
}

#[derive(PartialEq, Debug)]
pub struct Thumb {
    pub thumb_url: String,
    pub from: ThumbOf,
}

#[derive(PartialEq, Debug)]
pub struct Embed {
    pub embed_url: String,
    pub from: EmbedOf,
}

#[derive(PartialEq, Debug)]
pub enum Markup<'a> {
    Bold(Box<Markup<'a>>),
    Italic(Box<Markup<'a>>),
    Mono(Box<Markup<'a>>),
    Strike(Box<Markup<'a>>),
    Spoiler(Box<Markup<'a>>),
    Tex(&'a str),
    Code(Code<'a>),
    Quote(Box<Markup<'a>>),
    PostRef(Uuid),
    Link(Link<'a>),
    Newline,
    DoubleNewline,
    PlainText(&'a str),
    PlainPrefix(&'a str, Box<Markup<'a>>),
    Many(Vec<Markup<'a>>),
}

type In<'a> = (&'a str, u8);

#[derive(Debug, PartialEq)]
pub struct MarkupParseError<'a> {
    s: &'a str,
    kind: ErrorKind,
}

impl<'a> ParseError<In<'a>> for MarkupParseError<'a> {
    fn from_error_kind((input, _level): In<'a>, kind: ErrorKind) -> Self {
        Self { s: input, kind }
    }

    fn append((_input, _level): In<'a>, _kind: ErrorKind, other: Self) -> Self {
        other
    }
}

impl<'a> ParseError<&'a str> for MarkupParseError<'a> {
    fn from_error_kind(input: &'a str, kind: ErrorKind) -> Self {
        Self { s: input, kind }
    }

    fn append(_input: &'a str, _kind: ErrorKind, other: Self) -> Self {
        other
    }
}

type ParseResult<'a> = IResult<In<'a>, Markup<'a>, MarkupParseError<'a>>;


pub enum SimpleMarkup {
    Bold,
    Italic,
    Mono,
    Strike,
    Spoiler,
}

impl SimpleMarkup {
    pub fn parse(self) -> impl FnOnce(In) -> ParseResult {
        move |(s, level)| {
            let t = self.get_tag();

            let (r, m) = delimited(
                tag(t),
                take_until(t),
                tag(t),
            )(s)?;

            anychar(m)?;

            let (_, m) = parse_markup((m, level + 1))?;
            let m = self.to_markup(Box::new(m));

            Ok(((r, level), m))
        }
    }

    fn to_markup<'a, 'b>(&'a self, m: Box<Markup<'b>>) -> Markup<'b> {
        match self {
            SimpleMarkup::Bold => Markup::Bold(m),
            SimpleMarkup::Italic => Markup::Italic(m),
            SimpleMarkup::Mono => Markup::Mono(m),
            SimpleMarkup::Strike => Markup::Strike(m),
            SimpleMarkup::Spoiler => Markup::Spoiler(m),
        }
    }

    fn get_tag(&self) -> &'static str {
        match self {
            SimpleMarkup::Bold => "**",
            SimpleMarkup::Italic => "//",
            SimpleMarkup::Mono => "`",
            SimpleMarkup::Strike => "~~",
            SimpleMarkup::Spoiler => "%%",
        }
    }
}

pub fn parse_bold(s: In) -> ParseResult {
    SimpleMarkup::Bold.parse()(s)
}

pub fn parse_italic(s: In) -> ParseResult {
    SimpleMarkup::Italic.parse()(s)
}

pub fn parse_mono(s: In) -> ParseResult {
    SimpleMarkup::Mono.parse()(s)
}

pub fn parse_strike(s: In) -> ParseResult {
    SimpleMarkup::Strike.parse()(s)
}

pub fn parse_spoiler(s: In) -> ParseResult {
    SimpleMarkup::Spoiler.parse()(s)
}

pub fn parse_tex((s, level): In) -> ParseResult {
    let t = "$";

    let (r, m) = delimited(
        tag(t),
        take_until(t),
        tag(t),
    )(s)?;

    Ok(((r, level), Markup::Tex(m)))
}

pub fn parse_newline((s, level): In) -> ParseResult {
    let (r, m) = preceded(
        newline,
        multispace0,
    )(s)?;

    let m = if m.contains('\n') {
        Markup::DoubleNewline
    } else {
        Markup::Newline
    };

    Ok(((r, level), m))
}

fn is_whitespace(c: char) -> bool {
    c == ' ' || c == '\t' || c == '\r' || c == '\n'
}

fn parse_code((s, level): In) -> ParseResult {
    let t = "```";

    let (r, m) = delimited(
        tag(t),
        take_until(t),
        tag(t),
    )(s)?;

    let (code, lang) = terminated(
        take_till(is_whitespace),
        multispace1,
    )(m)?;

    // make it fail on empty string
    anychar(code)?;

    Ok(((r, level), Markup::Code(Code { lang, code })))
}

fn parse_uuid<'a, E: ParseError<&'a str>>(s: &'a str) -> IResult<&'a str, Uuid, E> {
    let (r, s) = alt((
        take_while_m_n(32, 32, |c: char| c.to_digit(16).is_some()),
        take_while_m_n(36, 36, |_| true)
    ))(s)?;

    let uuid = Uuid::parse_str(s).map_err(|_| {
        nom::Err::Error(E::from_error_kind(s, ErrorKind::IsNot))
    })?;

    Ok((r, uuid))
}

fn parse_ref((s, level): In) -> ParseResult {
    let (r, post_id) = preceded(tag("@"), parse_uuid)(s)?;

    Ok(((r, level), Markup::PostRef(post_id)))
}

fn parse_quote((s, level): In) -> ParseResult {
    // quotes are allowed only at the top level
    if level != 0 {
        return Err(nom::Err::Error(MarkupParseError { s, kind: ErrorKind::IsNot }));
    }

    let (r, m) = preceded(
        tag(">"),
        take_till(|c| c == '\n'),
    )(s)?;

    anychar(m)?;

    let (_, m) = parse_markup((m, level + 1))?;

    Ok(((r, level), Markup::Quote(Box::new(m))))
}

fn to_in_parser<'a>(str_parser: impl Fn(&'a str) -> IResult<&'a str, &'a str, MarkupParseError<'a>>)
                    -> impl Fn(In<'a>) -> IResult<In<'a>, &'a str, MarkupParseError<'a>>
{
    move |(s, l)| {
        let (r, m) = str_parser(s)?;
        Ok(((r, l), m))
    }
}

fn parse_quote_after_newline((s, level): In) -> ParseResult {
    let (r, (nl, q)) = tuple((
        parse_newline,
        preceded(to_in_parser(multispace0), parse_quote),
    ))((s, level))?;

    Ok((r, Markup::Many(vec![nl, q])))
}

fn parse_quote_after_ref((s, level): In) -> ParseResult {
    let (r, (post_ref, q)) = tuple((
        parse_ref,
        preceded(to_in_parser(multispace0), parse_quote),
    ))((s, level))?;

    Ok((r, Markup::Many(vec![post_ref, Markup::Newline, q])))
}

lazy_static! {
    static ref URL_RE: Regex = Regex::new(
        r"^(?:(?:https?|ftp)://(?:-\.)?(?:[^\s/?\.#]+\.?)+(?:/[^\s]*)?|magnet:\?(?:[^\s\]\)]+))"
    ).unwrap();
    static ref YT_URL_RE: Regex = Regex::new(
        r"^https?://(?:www\.|m\.)?(?:youtube\.com|youtu\.be)/(?:[\w\-]+\?v=|embed/|v/)?([\w\-]+)"
    ).unwrap();
    static ref IMGUR_URL_RE: Regex = Regex::new(
        r"^.*imgur.com/(\w{6,7})\w?\.(\w{3,4})"
    ).unwrap();
    static ref TWITTER_URL_RE: Regex = Regex::new(
        r"(^.*pbs.twimg.com/media/[\w_\-]+)(?:\.(\w{3,4})|\?format=(\w{3,4}))"
    ).unwrap();
    static ref PLOUDFR_URL_RE: Regex = Regex::new(
        r"^https?://(?:www\.)?video.ploud.fr/videos/(?:watch|embed)/([\w\-]+)"
    ).unwrap();
    static ref LIBREVIDEO_URL_RE: Regex = Regex::new(
        r"^https?://(?:www\.)?libre.video/videos/(?:watch|embed)/([\w\-]+)"
    ).unwrap();
    static ref PEERTUBESOCIAL_URL_RE: Regex = Regex::new(
        r"^https?://(?:www\.)?peertube.social/videos/(?:watch|embed)/([\w\-]+)"
    ).unwrap();
    static ref BITCHUTE_URL_RE: Regex = Regex::new(
        r"^https?://(?:www\.)?bitchute.com/(?:video|embed|torrent/[^/]+)/(\w+)"
    ).unwrap();
}

fn parse_link((s, level): In) -> ParseResult {
    match URL_RE.find(s) {
        Some(m) => {
            let url = &s[m.start()..m.end()];
            let r = &s[m.end()..];

            Ok(((r, level), Markup::Link(Link {
                url,
                name: None,
                media: parse_youtube_link(url)
                    .or_else(|| parse_librevideo_link(url))
                    .or_else(|| parse_peertubesocial_link(url))
                    .or_else(|| parse_ploudfr_link(url))
                    .or_else(|| parse_imgur_link(url))
                    .or_else(|| parse_twitter_link(url))
                    .map(Media::Thumb)
//                    .or_else(|| parse_bitchute_link(url).map(Media::Embed)),
            })))
        }
        None => {
            Err(nom::Err::Error(MarkupParseError { s, kind: ErrorKind::RegexpMatch }))
        }
    }
}

fn parse_youtube_link(url: &str) -> Option<Thumb> {
    let caps: Option<regex::Captures> = YT_URL_RE.captures(url);

    caps.and_then(|caps| caps.get(1).map(|m| {
        let video_id = m.as_str();
        Thumb {
            thumb_url: format!("https://img.youtube.com/vi/{}/mqdefault.jpg", video_id),
            from: ThumbOf::Youtube,
        }
    }))
}

#[allow(dead_code)]
fn parse_bitchute_link(url: &str) -> Option<Embed> {
    let caps: Option<regex::Captures> = BITCHUTE_URL_RE.captures(url);

    caps.and_then(|caps| caps.get(1).map(|m| {
        let video_id = m.as_str();
        Embed {
            embed_url: format!("https://www.bitchute.com/embed/{}", video_id),
            from: EmbedOf::Bitchute,
        }
    }))
}

fn parse_ploudfr_link(url: &str) -> Option<Thumb> {
    let caps: Option<regex::Captures> = PLOUDFR_URL_RE.captures(url);

    caps.and_then(|caps| caps.get(1).map(|m| {
        let video_id = m.as_str();
        Thumb {
            thumb_url: format!("https://video.ploud.fr/static/thumbnails/{}.jpg", video_id),
            from: ThumbOf::Ploudfr,
        }
    }))
}

fn parse_librevideo_link(url: &str) -> Option<Thumb> {
    let caps: Option<regex::Captures> = LIBREVIDEO_URL_RE.captures(url);

    caps.and_then(|caps| caps.get(1).map(|m| {
        let video_id = m.as_str();
        Thumb {
            thumb_url: format!("https://libre.video/static/thumbnails/{}.jpg", video_id),
            from: ThumbOf::LibreVideo,
        }
    }))
}

fn parse_peertubesocial_link(url: &str) -> Option<Thumb> {
    let caps: Option<regex::Captures> = PEERTUBESOCIAL_URL_RE.captures(url);

    caps.and_then(|caps| caps.get(1).map(|m| {
        let video_id = m.as_str();
        Thumb {
            thumb_url: format!("https://peertube.social/static/thumbnails/{}.jpg", video_id),
            from: ThumbOf::PeertubeSocial,
        }
    }))
}

fn parse_imgur_link(url: &str) -> Option<Thumb> {
    let caps: Option<regex::Captures> = IMGUR_URL_RE.captures(url);

    caps.and_then(|caps| zip_option(caps.get(1), caps.get(2)))
        .map(|(image_id, image_ext)| Thumb {
            thumb_url: format!("https://i.imgur.com/{}m.{}", image_id.as_str(), image_ext.as_str()),
            from: ThumbOf::Imgur,
        })
}

pub fn zip_option<T, U>(x: Option<T>, y: Option<U>) -> Option<(T, U)>
{
    Some((x?, y?))
}

pub fn zip_option3<T, U, R>(x: Option<T>, y: Option<U>, z: Option<R>) -> Option<(T, U, R)>
{
    Some((x?, y?, z?))
}

fn parse_twitter_link(url: &str) -> Option<Thumb> {
    let caps: Option<regex::Captures> = TWITTER_URL_RE.captures(url);

    caps.and_then(|caps| zip_option(caps.get(1), caps.get(2).or(caps.get(3)))
        .map(|(body, image_ext)| Thumb {
            thumb_url: format!("{}.{}:thumb", body.as_str(), image_ext.as_str()),
            from: ThumbOf::Twitter,
        })
    )
}

fn parse_named_link((s, level): In) -> ParseResult {
    let (r, name) = delimited(
        tag("["),
        take_until("]"),
        tag("]"),
    )(s)?;

    let (r, url) = delimited(
        tag("("),
        take_until(")"),
        tag(")"),
    )(r)?;

    Ok(((r, level), Markup::Link(Link { name: Some(name), url, media: None })))

//    let ((must_be_empty, _), link) = parse_link((m, level))?;
//
//    if must_be_empty != "" {
//        return Err(nom::Err::Error(MarkupParseError { s, kind: ErrorKind::RegexpMatch }));
//    }
//
//    match link {
//        Markup::Link(Link { name: _, url, thumb: _ }) => {
//            Ok(((r, level), Markup::Link(Link { name: Some(name), url, thumb: None })))
//        }
//        _ => panic!("it must be a link!"),
//    }
}

fn parse_markup_step((s, level): In) -> ParseResult {
    // make it fail on empty string
    anychar(s)?;

    if level >= 2 {
        return Ok((("", level), Markup::PlainText(s)));
    }

    for i in 0..s.len() {
        if !s.is_char_boundary(i) {
            continue;
        }

        let x = alt((
            parse_quote_after_newline,
            parse_newline,
            parse_bold,
            parse_italic,
            parse_mono,
            parse_strike,
            parse_tex,
            parse_spoiler,
            parse_code,
            parse_quote_after_ref,
            parse_ref,
            parse_link,
            parse_named_link,
        ))((&s[i..], level));

        if let Ok((r, m)) = x {
            assert_eq!(r.1, level);

            return if i == 0 {
                Ok((r, m))
            } else {
                Ok((r, Markup::PlainPrefix(&s[0..i], Box::new(m))))
            };
        }
    }

    Ok((("", level), Markup::PlainText(s)))
}

fn parse_markup((s, level): In) -> ParseResult {
    let (r, ms) = many0(parse_markup_step)((s, level))?;

    Ok((r, Markup::Many(ms)))
}

impl<'a> Markup<'a> {
    pub fn parse_to_html(s: &'a str) -> String {
        let m = Self::parse(s).map_err(|err|
            error!("Failed to parse a post: {}", err)
        ).unwrap_or_else(|_| Markup::PlainText(s));

        m.to_html()
    }

    fn parse(s: &'a str) -> Result<Self, nom::Err<MarkupParseError>> {
        let s = s.trim_start();
        // there might be a quote right at the beginning
        let (r, m) = match parse_quote((s, 0)) {
            Ok((r, q)) => {
                let (r, m) = parse_markup(r)?;
                (r, Markup::Many(vec![q, m]))
            }
            _ => parse_markup((s, 0))?
        };
        assert_eq!(r, ("", 0));
        Ok(m)
    }

    fn to_html(&self) -> String {
        let settings: &Settings = &crate::SETTINGS;

        match self {
            Markup::Bold(m) => format!("<b>{}</b>", m.to_html()),
            Markup::Italic(m) => format!("<i>{}</i>", m.to_html()),
            Markup::Mono(m) => format!("<samp>{}</samp>", m.to_html()),
            Markup::Strike(m) => format!("<s>{}</s>", m.to_html()),
            Markup::Spoiler(m) => format!("<span class=\"spoiler\">{}</span>", m.to_html()),
            Markup::Tex(t) => format!("<span class=\"tex\">${}$</span>", escape_text(t)),
            Markup::Code(Code { lang, code }) => {
                format!(
                    "<pre><code class=\"{}\">{}</code></pre>",
                    escape_attribute(lang),
                    escape_text(code),
                )
            }
            Markup::Quote(m) => {
                format!("<span class=\"quote\">&gt;{}</span>", m.to_html())
            }
            Markup::PostRef(post_id) => {
                let post_id_simple = post_id.to_simple().to_string();
                let short_post_id = &post_id_simple[0..5];
                format!(
                    "<span class=\"post_ref id\"><a href=\"{}/@/{}\">@{}</a>\
                    <span class=\"full_id_block\">\
                    <span class=\"hidden_text\">@</span>\
                    <span class=\"full_id\">{}</span></span></span>",
                    settings.ui.base_path,
                    post_id.to_simple(),
                    short_post_id,
                    post_id_simple,
                )
            }
            Markup::Link(Link { name, url, media }) => {
                let visible_url = escape_text(
                    name.unwrap_or(Self::decode_url(url).as_ref())
                );
                let href = Self::escape_href(url);

                match media {
                    None => format!(
                        "<a href=\"{}\" target=\"_blank\">{}</a>",
                        href, visible_url
                    ),
                    Some(Media::Thumb(thumb)) => {
                        let thumb_class = match thumb.from {
                            ThumbOf::Youtube => "youtube_thumb",
                            ThumbOf::Imgur => "imgur_thumb",
                            ThumbOf::Twitter => "twitter_thumb",
                            ThumbOf::LibreVideo => "librevideo_thumb",
                            ThumbOf::PeertubeSocial => "peertubesocial_thumb",
                            ThumbOf::Ploudfr => "ploudfr_thumb",
                        };

                        format!(
                            "<a href=\"{}\" target=\"_blank\" class=\"thumb {}\">\
                            <div class=\"image_parent\">\
                            <div class=\"image\" style=\"background-image: url('{}');\"></div>\
                            </div>\
                            <span>{}</span></a>",
                            href, thumb_class, Self::escape_href(&thumb.thumb_url), visible_url
                        )
                    },
                    Some(Media::Embed(embed)) => {
                        let embed_class = match embed.from {
                            EmbedOf::Bitchute => "bitchute_embed",
                        };

                        format!(
                            "<a href=\"{}\" target=\"_blank\" class=\"embed {}\">\
                            <iframe src=\"{}\" allowfullscreen></iframe>\
                            <br>{}</a>",
                            href, embed_class, Self::escape_href(&embed.embed_url), visible_url
                        )
                    },
                }
            }
            Markup::Newline => "<br>".to_string(),
            Markup::DoubleNewline => "<br><br>".to_string(),
            Markup::PlainText(text) => escape_text(text),
            Markup::PlainPrefix(text, m) => {
                format!("{}{}", escape_text(text), m.to_html())
            }
            Markup::Many(ms) => Self::many_to_html(ms),
        }
    }

    fn many_to_html(ms: &Vec<Markup>) -> String {
        ms.into_iter().map(|m| m.to_html())
            .collect::<Vec<String>>().join("")
    }

    const FRAGMENT: &'static AsciiSet = &CONTROLS.add(b'<').add(b'>');

    fn escape_href(url: &str) -> String {
        // escape_attribute will encode <> as &lt; &gt;, so to avoid it, we percent-encode them
        let url = utf8_percent_encode(url, Self::FRAGMENT).to_string();
        escape_attribute(url)
    }

    fn decode_url(url: &str) -> Cow<str> {
        percent_decode_str(url).decode_utf8_lossy()
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_parsing_simple_markup() {
        assert_eq!(Markup::parse_to_html("Фыва"), "Фыва");
        assert_eq!(Markup::parse_to_html("asdf"), "asdf");
        assert_eq!(Markup::parse_to_html("**asdf"), "**asdf");

        assert_eq!(
            Markup::parse_to_html(" **<b>** "),
            "<b>&lt;b&gt;</b> "
        );

        assert_eq!(
            Markup::parse_to_html("\t\n **b //<'i\">// ** "),
            "<b>b <i>&lt;'i\"&gt;</i> </b> "
        );

        assert_eq!(
            Markup::parse_to_html("asfd $**b**$"),
            "asfd <span class=\"tex\">$**b**$</span>"
        );

        assert_eq!(Markup::parse_to_html("asdf\nzxcv"), "asdf<br>zxcv");
        assert_eq!(Markup::parse_to_html("**asdf\n\nzxcv**"), "<b>asdf<br><br>zxcv</b>");
        assert_eq!(Markup::parse_to_html("asdf\n \t \nzxcv"), "asdf<br><br>zxcv");
        assert_eq!(Markup::parse_to_html("asdf\n \n \n \n  zxcv"), "asdf<br><br>zxcv");

        assert_eq!(
            Markup::parse_to_html("asfd %%b%%"),
            "asfd <span class=\"spoiler\">b</span>"
        );
    }

    #[test]
    fn test_parsing_code() {
        assert_eq!(
            Markup::parse_to_html(" ```py```"),
            "``<samp>py</samp>``"
        );

        assert_eq!(
            Markup::parse_to_html(" ```python \n ```"),
            "``<samp>python <br></samp>``"
        );

        assert_eq!(
            Markup::parse_to_html(" ```py\"thon 2+2=5``` "),
            "<pre><code class=\"py&quot;thon\">2+2=5</code></pre> "
        );

        assert_eq!(
            Markup::parse_to_html(" ```\n2+2=5```"),
            "<pre><code class=\"\">2+2=5</code></pre>"
        );
    }

    #[test]
    fn test_parsing_post_ref() {
        assert_eq!(
            Markup::parse_to_html("**b** @7105c424d0584affb1e9b3567f341582"),
            "<b>b</b> <span class=\"post_ref id\">\
            <a href=\"/ui3/@/7105c424d0584affb1e9b3567f341582\">@7105c</a>\
            <span class=\"full_id_block\">\
            <span class=\"hidden_text\">@</span>\
            <span class=\"full_id\">7105c424d0584affb1e9b3567f341582</span></span></span>"
        );
    }

    #[test]
    fn test_parsing_quotes() {
        assert_eq!(
            Markup::parse_to_html("> some quote"),
            "<span class=\"quote\">&gt; some quote</span>"
        );
        assert_eq!(
            Markup::parse_to_html(" \t > some quote"),
            "<span class=\"quote\">&gt; some quote</span>"
        );
        assert_eq!(
            Markup::parse_to_html(" \t > some quote"),
            "<span class=\"quote\">&gt; some quote</span>"
        );
        assert_eq!(
            Markup::parse_to_html(" \t > some >quote"),
            "<span class=\"quote\">&gt; some &gt;quote</span>"
        );
        assert_eq!(
            Markup::parse_to_html(" some >quote"),
            "some &gt;quote"
        );
        assert_eq!(
            Markup::parse_to_html(" some  \n \n    >quote"),
            "some  <br><br><span class=\"quote\">&gt;quote</span>"
        );
        assert_eq!(
            Markup::parse_to_html(" 11@7105c424d0584affb1e9b3567f341582 > some quote"),
            "11<span class=\"post_ref id\"><a href=\"/ui3/@/7105c424d0584affb1e9b3567f341582\">\
            @7105c</a><span class=\"full_id_block\">\
            <span class=\"hidden_text\">@</span>\
            <span class=\"full_id\">7105c424d0584affb1e9b3567f341582</span>\
            </span></span>\
            <br>\
            <span class=\"quote\">&gt; some quote</span>"
        );
    }

    #[test]
    fn test_parsing_plain_links() {
        assert_eq!(
            Markup::parse_to_html("https://docs.rs/regex/1.3.9/regex/index.html asdfasdf."),
            "<a href=\"https://docs.rs/regex/1.3.9/regex/index.html\" \
            target=\"_blank\">\
            https://docs.rs/regex/1.3.9/regex/index.html</a> asdfasdf."
        );

        assert_eq!(
            Markup::parse_to_html("https://ru.wikipedia.org/wiki/%D0%AD%D0%B3%D0%BE"),
            "<a href=\"https://ru.wikipedia.org/wiki/%D0%AD%D0%B3%D0%BE\" \
            target=\"_blank\">\
            https://ru.wikipedia.org/wiki/Эго</a>"
        );

        assert_eq!(
            Markup::parse_to_html("**ftp://docs.rs/p=1&s=2\"></a>**"),
            "<b><a href=\"ftp://docs.rs/p=1&amp;s=2&quot;%3E%3C/a%3E\" \
            target=\"_blank\">\
            ftp://docs.rs/p=1&amp;s=2\"&gt;&lt;/a&gt;</a></b>"
        );

        assert_eq!(
            Markup::parse_to_html("(magnet:?xt=urn:btih:c12fe1c06bba254a9dc9f519b335aa7c1367a88a)"),
            "(<a href=\"magnet:?xt=urn:btih:c12fe1c06bba254a9dc9f519b335aa7c1367a88a\" \
            target=\"_blank\">\
            magnet:?xt=urn:btih:c12fe1c06bba254a9dc9f519b335aa7c1367a88a</a>)"
        );
    }

    #[test]
    fn test_parsing_media_links() {
        assert_eq!(
            Markup::parse_to_html("https://www.youtube.com/watch?v=dQw4w9WgXcQ"),
            "<a href=\"https://www.youtube.com/watch?v=dQw4w9WgXcQ\" \
            target=\"_blank\" class=\"thumb youtube_thumb\">\
            <div class=\"image_parent\">\
            <div class=\"image\" style=\"background-image: \
            url('https://img.youtube.com/vi/dQw4w9WgXcQ/mqdefault.jpg');\"></div>\
            </div>\
            <span>https://www.youtube.com/watch?v=dQw4w9WgXcQ</span>\
            </a>"
        );
        assert_eq!(
            Markup::parse_to_html("https://youtu.be/dQw4w9WgXcQ"),
            "<a href=\"https://youtu.be/dQw4w9WgXcQ\" \
            target=\"_blank\" class=\"thumb youtube_thumb\">\
            <div class=\"image_parent\">\
            <div class=\"image\" style=\"background-image: \
            url('https://img.youtube.com/vi/dQw4w9WgXcQ/mqdefault.jpg');\"></div>\
            </div>\
            <span>https://youtu.be/dQw4w9WgXcQ</span></a>"
        );
        assert_eq!(
            Markup::parse_to_html("https://i.imgur.com/JLMfo17.jpeg"),
            "<a href=\"https://i.imgur.com/JLMfo17.jpeg\" \
            target=\"_blank\" class=\"thumb imgur_thumb\">\
            <div class=\"image_parent\">\
            <div class=\"image\" style=\"background-image: \
            url('https://i.imgur.com/JLMfo17m.jpeg');\"></div>\
            </div>\
            <span>https://i.imgur.com/JLMfo17.jpeg</span></a>"
        );
        assert_eq!(
            Markup::parse_to_html("https://pbs.twimg.com/media/Eh-KoAfXsAAfwiu?format=jpg&name=medium"),
            "<a href=\"https://pbs.twimg.com/media/Eh-KoAfXsAAfwiu?format=jpg&amp;name=medium\" \
            target=\"_blank\" class=\"thumb twitter_thumb\">\
            <div class=\"image_parent\">\
            <div class=\"image\" style=\"background-image: \
            url('https://pbs.twimg.com/media/Eh-KoAfXsAAfwiu.jpg:thumb');\"></div>\
            </div>\
            <span>https://pbs.twimg.com/media/Eh-KoAfXsAAfwiu?format=jpg&amp;name=medium</span></a>"
        );
        assert_eq!(
            Markup::parse_to_html("https://pbs.twimg.com/media/Eh-KoAfXsAAfwiu.jpg"),
            "<a href=\"https://pbs.twimg.com/media/Eh-KoAfXsAAfwiu.jpg\" \
            target=\"_blank\" class=\"thumb twitter_thumb\">\
            <div class=\"image_parent\">\
            <div class=\"image\" style=\"background-image: \
            url('https://pbs.twimg.com/media/Eh-KoAfXsAAfwiu.jpg:thumb');\"></div>\
            </div>\
            <span>https://pbs.twimg.com/media/Eh-KoAfXsAAfwiu.jpg</span></a>"
        );
        assert_eq!(
            Markup::parse_to_html("https://libre.video/videos/watch/b985c5ca-fdcf-46ff-92d5-e68922fe4821"),
            "<a href=\"https://libre.video/videos/watch/b985c5ca-fdcf-46ff-92d5-e68922fe4821\" \
            target=\"_blank\" class=\"thumb librevideo_thumb\">\
            <div class=\"image_parent\">\
            <div class=\"image\" style=\"background-image: \
            url('https://libre.video/static/thumbnails/b985c5ca-fdcf-46ff-92d5-e68922fe4821.jpg');\"></div>\
            </div>\
            <span>https://libre.video/videos/watch/b985c5ca-fdcf-46ff-92d5-e68922fe4821</span>\
            </a>"
        );
        assert_eq!(
            Markup::parse_to_html("https://peertube.social/videos/watch/b985c5ca-fdcf-46ff-92d5-e68922fe4821"),
            "<a href=\"https://peertube.social/videos/watch/b985c5ca-fdcf-46ff-92d5-e68922fe4821\" \
            target=\"_blank\" class=\"thumb peertubesocial_thumb\">\
            <div class=\"image_parent\">\
            <div class=\"image\" style=\"background-image: \
            url('https://peertube.social/static/thumbnails/b985c5ca-fdcf-46ff-92d5-e68922fe4821.jpg');\"></div>\
            </div>\
            <span>https://peertube.social/videos/watch/b985c5ca-fdcf-46ff-92d5-e68922fe4821</span>\
            </a>"
        );
        assert_eq!(
            Markup::parse_to_html("https://video.ploud.fr/videos/watch/b985c5ca-fdcf-46ff-92d5-e68922fe4821"),
            "<a href=\"https://video.ploud.fr/videos/watch/b985c5ca-fdcf-46ff-92d5-e68922fe4821\" \
            target=\"_blank\" class=\"thumb ploudfr_thumb\">\
            <div class=\"image_parent\">\
            <div class=\"image\" style=\"background-image: \
            url('https://video.ploud.fr/static/thumbnails/b985c5ca-fdcf-46ff-92d5-e68922fe4821.jpg');\"></div>\
            </div>\
            <span>https://video.ploud.fr/videos/watch/b985c5ca-fdcf-46ff-92d5-e68922fe4821</span>\
            </a>"
        );
//        assert_eq!(
//            Markup::parse_to_html("https://www.bitchute.com/video/FjFl3ZH0Xwuw/"),
//            "<a href=\"https://www.bitchute.com/video/FjFl3ZH0Xwuw/\" target=\"_blank\" class=\"embed bitchute_embed\">\
//            <iframe src=\"https://www.bitchute.com/embed/FjFl3ZH0Xwuw\" allowfullscreen></iframe>\
//            <br>https://www.bitchute.com/video/FjFl3ZH0Xwuw/</a>"
//        );
    }

    #[test]
    fn test_parsing_named_links() {
        assert_eq!(
            Markup::parse_to_html("[na<me](https://docs.rs/regex) asdfasdf."),
            "<a href=\"https://docs.rs/regex\" target=\"_blank\">\
            na&lt;me</a> asdfasdf."
        );

        assert_eq!(
            Markup::parse_to_html("[name](asdfasdf)."),
            "<a href=\"asdfasdf\" target=\"_blank\">name</a>."
        );

        assert_eq!(
            Markup::parse_to_html("[na<me](https://docs.rs/regex asdfasdf)."),
            "<a href=\"https://docs.rs/regex asdfasdf\" target=\"_blank\">na&lt;me</a>."
        );

        // do not convert named links to thumbs
        assert_eq!(
            Markup::parse_to_html("[name](https://www.youtube.com/watch?v=dQw4w9WgXcQ)"),
            "<a href=\"https://www.youtube.com/watch?v=dQw4w9WgXcQ\" \
            target=\"_blank\">name</a>"
        );
    }
}