use warp::http;
use warp::Filter;
use warp::multipart::{FormData};
use warp::filters::BoxedFilter;
use std::str::FromStr;
use crate::models::{Board, NewPost, Cookies, UserSettings, UpdatedHiddenByOp, Auth, NewUser};
use uuid::Uuid;
use serde_derive::Deserialize;
use crate::pages;
use crate::db_service::open_db;
use std::sync::Arc;

pub fn get_routes() -> BoxedFilter<(impl warp::Reply,)> {
    let db: sled::Db = open_db(&crate::SETTINGS.database).unwrap();
    let db = Arc::new(db);
    let db = warp::any().map(move || db.clone());

    let client_id_cookie = warp::cookie::optional("clientId");
    let access_token_cookie = warp::cookie::optional("accessToken");

    let cookies = warp::any()
        .and(client_id_cookie)
        .and(access_token_cookie)
        .and(warp::cookie::optional("fixMenu"))
        .and(warp::cookie::optional("showThumbs"))
        .and(warp::cookie::optional("maxWidth"))
        .and(warp::cookie::optional("compact"))
        .and(warp::cookie::optional("onlyChatPosts"))
        .map(|client_id: Option<String>, access_token: Option<String>, fix_menu: Option<String>,
              show_thumbs: Option<String>, max_width: Option<String>, compact: Option<String>,
              only_chat_posts: Option<String>|
            Cookies {
                client_id,
                access_token,
                fix_menu: fix_menu.and_then(|x| x.parse::<bool>().ok())
                    .unwrap_or(crate::SETTINGS.ui.fix_menu),
                show_thumbs: show_thumbs.and_then(|x| x.parse::<bool>().ok())
                    .unwrap_or(crate::SETTINGS.ui.show_thumbs),
                max_width: max_width.and_then(|x| x.parse::<u32>().ok())
                    .unwrap_or(crate::SETTINGS.ui.max_width),
                compact: compact.and_then(|x| x.parse::<bool>().ok())
                    .unwrap_or(crate::SETTINGS.ui.compact),
                only_chat_posts: only_chat_posts.and_then(|x| x.parse::<bool>().ok())
                    .unwrap_or(crate::SETTINGS.ui.only_chat_posts),
            });

    let referer_header = warp::header::optional::<String>(http::header::REFERER.as_str());
    let if_none_match = warp::header::optional::<String>(http::header::IF_NONE_MATCH.as_str());

    let warp_route = warp::path::end()
        .and(warp::get())
        .and(warp::query::query())
        .and(cookies)
        .and(if_none_match)
        .and(db.clone())
        .and_then(|page: Page, cookies: Cookies, req_et: Option<String>, db: Arc<sled::Db>|
            async move {
                result(pages::warp_page(None, page.p, cookies, req_et, db).await)
            })
        .boxed();

    let board_route = warp::get()
        .and(warp::path::param())
        .and(warp::path::end())
        .and(warp::query::query())
        .and(cookies)
        .and(if_none_match)
        .and(db.clone())
        .and_then(|board_path: BoardPath, page: Page, cookies: Cookies,
                   req_et: Option<String>, db: Arc<sled::Db>|
            async move {
                result(pages::warp_page(Some(board_path.0), page.p, cookies, req_et, db).await)
            })
        .boxed();

    let thread_route = warp::get()
        .and(warp::path::param())
        .and(warp::path::param())
        .and(warp::path::end())
        .and(warp::query::query())
        .and(cookies)
        .and(if_none_match)
        .and(db.clone())
        .and_then(|_: BoardPath, thread_id: Uuid, page: Page, cookies: Cookies,
                   req_et: Option<String>, db: Arc<sled::Db>|
            async move {
                result(pages::thread_page(thread_id, page.p, cookies, req_et, db).await)
            })
        .boxed();

    let post_route = warp::path("@")
        .and(warp::get())
        .and(warp::path::param())
        .and(warp::path::end())
        .and_then(|post_id: Uuid|
            async move { result(pages::post_page(post_id).await) })
        .boxed();

    let chat_route = warp::path("chat")
        .and(warp::get())
        .and(warp::path::end())
        .and(cookies)
        .and(if_none_match)
        .and_then(|cookies: Cookies, req_et: Option<String>|
            async move { result(pages::chat_page(cookies, req_et).await) })
        .boxed();

    let chat_frame_route = warp::path("chatFrame")
        .and(warp::get())
        .and(warp::path::end())
        .and(cookies)
        .and(if_none_match)
        .and(db.clone())
        .and_then(|cookies: Cookies, req_et: Option<String>, db: Arc<sled::Db>|
            async move { result(pages::chat_frame(cookies, req_et, db).await) })
        .boxed();

    let boards_route = warp::path("boards")
        .and(warp::get())
        .and(warp::path::end())
        .and(warp::query::query())
        .and(cookies)
        .and(if_none_match)
        .and_then(|page: Page, cookies: Cookies, req_et: Option<String>|
            async move { result(pages::boards_page(page.p, cookies, req_et).await) })
        .boxed();

    let help_route = warp::path("help")
        .and(warp::get())
        .and(warp::path::end())
        .and(cookies)
        .and(if_none_match)
        .and_then(|cookies: Cookies, req_et: Option<String>|
            async move { result(pages::help_page(cookies, req_et).await) })
        .boxed();

    let export_route = warp::path("pfpmd_ui3_settings.json")
        .and(warp::get())
        .and(warp::path::end())
        .and(cookies)
        .and(if_none_match)
        .and_then(|cookies: Cookies, req_et: Option<String>|
            async move { result(pages::export(cookies, req_et).await) })
        .boxed();

    let import_route = warp::path("import")
        .and(warp::get())
        .and(warp::path::end())
        .and(cookies)
        .and(if_none_match)
        .and(referer_header)
        .and_then(|cookies: Cookies, req_et: Option<String>, referer: Option<String>|
            async move { result(pages::import_page(cookies, req_et, referer).await) })
        .boxed();

    let settings_route = warp::path("settings")
        .and(warp::get())
        .and(warp::path::end())
        .and(cookies)
        .and(if_none_match)
        .and(referer_header)
        .and_then(|cookies: Cookies, req_et: Option<String>, referer: Option<String>|
            async move { result(pages::settings_page(cookies, req_et, referer).await) })
        .boxed();

    let login_route = warp::path("login")
        .and(warp::get())
        .and(warp::path::end())
        .and(cookies)
        .and(if_none_match)
        .and(referer_header)
        .and_then(|cookies: Cookies, req_et: Option<String>, referer: Option<String>|
            async move { result(pages::login_page(cookies, req_et, referer).await) })
        .boxed();

    let profile_route = warp::path("profile")
        .and(warp::get())
        .and(warp::path::end())
        .and(cookies)
        .and(if_none_match)
        .and(referer_header)
        .and_then(|cookies: Cookies, req_et: Option<String>, referer: Option<String>|
            async move { result(pages::profile_page(cookies, req_et, referer).await) })
        .boxed();

    let publish_route = warp::path("publish")
        .and(warp::post())
        .and(warp::path::end())
        .and(warp::body::content_length_limit(1024 * 16))
        .and(warp::body::form())
        .and(cookies)
        .and(referer_header)
        .and_then(|new_post: NewPost, cookies: Cookies, referer: Option<String>|
            async move { result(pages::publish(new_post, cookies, referer).await) })
        .boxed();

    let save_settings_route = warp::path("saveSettings")
        .and(warp::post())
        .and(warp::path::end())
        .and(warp::body::content_length_limit(1024))
        .and(warp::body::form())
        .and(cookies)
        .and_then(|user_settings: UserSettings, cookies: Cookies|
            async move { result(pages::save_settings(user_settings, cookies).await) })
        .boxed();

    let authorize_route = warp::path("authorize")
        .and(warp::post())
        .and(warp::path::end())
        .and(warp::body::content_length_limit(256))
        .and(warp::body::form())
        .and(cookies)
        .and_then(|auth: Auth, cookies: Cookies|
            async move { result(pages::authorize(auth, cookies).await) })
        .boxed();

    let logout_route = warp::path("logout")
        .and(warp::post())
        .and(warp::path::end())
        .and(warp::body::content_length_limit(256))
        .and(cookies)
        .and_then(|cookies: Cookies|
            async move { result(pages::logout(cookies).await) })
        .boxed();

    let signup_route = warp::path("signup")
        .and(warp::post())
        .and(warp::path::end())
        .and(warp::body::content_length_limit(1024))
        .and(warp::body::form())
        .and(cookies)
        .and_then(|new_user: NewUser, cookies: Cookies|
            async move { result(pages::signup(new_user, cookies).await) })
        .boxed();

    let update_hidden_by_op_route = warp::path("updateHiddenByOp")
        .and(warp::post())
        .and(warp::path::end())
        .and(warp::body::content_length_limit(1024))
        .and(warp::body::form())
        .and(cookies)
        .and(referer_header)
        .and_then(|updated_hidden_by_op: UpdatedHiddenByOp, cookies: Cookies, referer: Option<String>|
            async move { result(pages::update_hidden_by_op(updated_hidden_by_op, cookies, referer).await) })
        .boxed();

    let import_ui_state_route = warp::path("importUiState")
        .and(warp::post())
        .and(warp::path::end())
        .and(warp::body::content_length_limit(1024 * 2))
        .and(warp::multipart::form().max_length(1024 * 2))
        .and(cookies)
        .and_then(|form: FormData, cookies: Cookies|
            async move { result(pages::import_ui_state(form, cookies).await)
        })
        .boxed();

    // .boxed() was added because of this https://github.com/seanmonstar/warp/issues/811
    warp_route
        .or(board_route)
        .or(thread_route)
        .boxed()
        .or(post_route)
        .or(chat_route)
        .boxed()
        .or(chat_frame_route)
        .boxed()
        .or(login_route)
        .boxed()
        .or(profile_route)
        .boxed()
        .or(authorize_route)
        .boxed()
        .or(signup_route)
        .boxed()
        .or(logout_route)
        .boxed()
        .or(boards_route)
        .or(help_route)
        .boxed()
        .or(settings_route)
        .boxed()
        .or(import_route)
        .or(export_route)
        .boxed()
        .or(publish_route)
        .boxed()
        .or(save_settings_route)
        .or(update_hidden_by_op_route)
        .or(import_ui_state_route)
        .boxed()
        .with(warp::reply::with::header(http::header::REFERRER_POLICY, "same-origin"))
        .boxed()
}

pub fn result(r: Result<warp::reply::Response, warp::reply::Response>)
              -> Result<warp::reply::Response, warp::reject::Rejection>
{
    match r {
        Ok(r) => Ok(r),
        Err(r) => Ok(r),
    }
}

struct BoardPath(String);

impl FromStr for BoardPath {
    type Err = ();
    fn from_str(board_name: &str) -> Result<BoardPath, ()> {
        let board_name = board_name.to_string();

        if Board::is_valid_name(&board_name) {
            Ok(BoardPath(board_name))
        } else {
            Err(())
        }
    }
}

#[derive(Deserialize)]
struct Page {
    p: Option<u32>
}
