use std::env;
use config::{Config, File, Environment};
use serde_derive::Deserialize;
use std::net::IpAddr;

#[derive(Debug, Deserialize)]
pub struct Settings {
    pub server: Server,
    pub api_server: ApiServer,
    pub api_client: ApiClientSettings,
    pub ui: UiSettings,
    pub database: DbSettings,
}

/// Server settings.
#[derive(Debug, Deserialize)]
pub struct Server {
    pub port: u16,
    pub host: IpAddr,
}

/// API server settings.
#[derive(Debug, Deserialize)]
pub struct ApiServer {
    pub url: String,
}

/// API client settings.
#[derive(Debug, Deserialize)]
pub struct ApiClientSettings {
    pub request_timeout: u64,
    pub proxy: Option<String>,
}

/// UI settings.
#[derive(Debug, Deserialize)]
pub struct UiSettings {
    pub warp_page_size: u32,
    pub page_size: u32,
    pub base_path: String,
    pub pagination_sibling_count: u32,
    pub fix_menu: bool,
    pub show_thumbs: bool,
    pub max_width: u32,
    pub compact: bool,
    pub only_chat_posts: bool,
}

// DB settings
#[derive(Debug, Deserialize)]
pub struct DbSettings {
    pub path: String,
    pub cache_capacity: u64,
    pub use_compression: bool,
    pub compression_factor: i32,
//    pub temporary: bool,
}

impl Settings {
    pub fn new() -> Result<Self, Box<dyn std::error::Error>> {
        let mut s = Config::new();

        s.merge(File::with_name("config/default"))?;

        // Add in the current environment file
        // Default to 'development' env
        // Note that this file is _optional_
        let env = env::var("RUN_MODE").unwrap_or("development".into());
        s.merge(File::with_name(&format!("config/{}", env)).required(false))?;

        // Add in settings from the environment (with a prefix of PFPMD)
        s.merge(Environment::with_prefix("pfpmd"))?;

        Ok(s.try_into()?)
    }
}