from shared import *
from test_authorize import test_authorize_success
from test_board_page import get_board_page
from test_boards_page import get_boards_page
from test_thread_page import get_thread_page


def publish(data, params=None, referrer=None, clear_cookies=True, cookies=None):
    if clear_cookies:
        session.cookies.clear()

    headers = {'Referer': referrer} if referrer else None

    r = session.post(f"{base_url}/publish", data=data, params=params,
                     headers=headers, allow_redirects=False, cookies=cookies)
    print(f"POST {r.url}\n{data}")
    return r


def test_publish():
    access_token = test_authorize_success().cookies.get('accessToken')
    redirect_url = f"{base_path}/help"

    # even if post wasn't created, no error is returned
    # here actually clientId is missing
    r = publish(data={'text': 'asdf'}, referrer=redirect_url, cookies={'accessToken': access_token})
    assert_found(r, redirect_url)

    # post to chat
    r = publish(data={'text': 'test123', 'threadId': chat_thread_id, 'some': 'thing'},
                referrer=redirect_url,
                cookies={'clientId': client_id, 'accessToken': access_token})
    assert_found(r, redirect_url_prefix=f'{base_path}/cht/{chat_thread_id}#')

    # test back_url
    r = publish(data={'text': 'test123', 'threadId': chat_thread_id,
                      'backUrl': '/chat', 'stayHere': 'true'},
                referrer=redirect_url,
                cookies={'clientId': client_id, 'accessToken': access_token})
    assert_found(r, redirect_url='/chat')

    # create new thread
    r = publish(data={'text': 'test123', 'boardName': 'b', 'subject': 'aaa'},
                referrer=redirect_url,
                cookies={'clientId': client_id, 'accessToken': access_token})
    assert_found(r, redirect_url_prefix=f'{base_path}/b/')

    # `referer` is not required
    r = publish(data={'text': 'asdf2', 'threadId': chat_thread_id},
                cookies={'accessToken': access_token})
    assert_found(r, f"{base_path}/")


def test_publish_cli():
    access_token = test_authorize_success().cookies.get('accessToken')

    r = publish(data={'text': '  /help\n', 'threadId': chat_thread_id},
                referrer=f'{base_path}/',
                cookies={'clientId': client_id})
    assert_found(r, redirect_url=f'{base_path}/help')

    r = publish(data={'text': '\n  /settings\n', 'threadId': chat_thread_id},
                referrer=f'{base_path}/',
                cookies={'clientId': client_id})
    assert_found(r, redirect_url=f'{base_path}/settings')

    r = publish(data={'text': '  /b  test 123description  ', 'threadId': chat_thread_id},
                cookies={'clientId': client_id, 'accessToken': access_token})
    assert_found(r)
    r = get_boards_page()
    assert 'test 123description' in r.text
    publish(data={'text': '  /b random', 'threadId': chat_thread_id},
            cookies={'clientId': client_id, 'accessToken': access_token})

    import random
    text = f'test text {random.random()}'
    r = publish(data={'text': f'  /b/e2ba4f60c472403ab26e7cfd749033d0 {text} ',
                      'threadId': chat_thread_id},
                cookies={'clientId': client_id, 'accessToken': access_token})
    assert_found(r)
    r = get_thread_page('b', 'e2ba4f60c472403ab26e7cfd749033d0')
    assert text in r.text

    subject = f'test subject {random.random()}'
    text = f'test text {random.random()}'
    r = publish(data={'text': f'  /b[{subject}]  {text}  ', 'threadId': chat_thread_id},
                cookies={'clientId': client_id, 'accessToken': access_token})
    assert_found(r)
    r = get_board_page('b')
    assert subject in r.text
    assert text in r.text

    r = publish(data={'text': '  /cc ', 'threadId': chat_thread_id},
                cookies={'clientId': client_id, 'accessToken': access_token})
    assert r.cookies['clientId'] != ""
    assert r.cookies['clientId'] != client_id


def test_publish_errors():
    access_token = test_authorize_success().cookies.get('accessToken')

    # `text` required
    r = publish(data={'something': 'true'})
    assert r.text == 'Request body deserialize error: missing field `text`'
    assert r.status_code == 400

    redirect_url = f"{base_path}/help"
    for thread_id in invalid_uuids:
        r = publish(data={'text': 'test123', 'threadId': thread_id},
                    referrer=redirect_url, cookies={'clientId': client_id})
        assert r.status_code == 400

    # unauthorized
    r = publish(data={'text': 'test123', 'board': 'b'},
                referrer=redirect_url, cookies={'clientId': client_id})
    assert_unauthorized(r)

    # missing field name: `board` instead of `board_name`
    r = publish(data={'text': 'test123', 'board': 'b', 'subject': 'test'},
                referrer=redirect_url,
                cookies={'clientId': client_id, 'accessToken': access_token})
    assert r.status_code == 200
    assert 'boardName is required' in r.text

    # subject missing
    r = publish(data={'text': 'test123', 'boardName': 'b'},
                referrer=redirect_url,
                cookies={'clientId': client_id, 'accessToken': access_token})
    assert r.status_code == 200
    assert 'Cannot create a thread without a subject' in r.text

    for board_name in invalid_boards:
        r = publish(data={'text': 'test123', 'boardName': board_name},
                    referrer=redirect_url,
                    cookies={'clientId': client_id, 'accessToken': access_token})
        assert r.status_code == 200
        assert 'Invalid board name' in r.text

    # max length
    r = publish(data={'something': 'a' * (1024 * 16)})
    assert r.status_code == 413

    r = session.get(f"{base_url}/publish")
    print(f"GET {r.url}")
    assert r.status_code == 405
